package com.jdblabs.jlp

import com.jdblabs.jlp.ast.*
import com.jdblabs.jlp.ast.Directive.DirectiveType

import org.pegdown.Extensions
import org.pegdown.PegDownProcessor

import java.util.List

public class LiterateMarkdownGenerator extends JLPBaseGenerator {

    protected PegDownProcessor pegdown

    public LiterateMarkdownGenerator(Processor processor) {
        super(processor)

        pegdown = new PegDownProcessor(
            Extensions.TABLES | Extensions.DEFINITIONS) }

    protected void parse(Directive directive) {
        switch(directive.type) {
            case DirectiveType.Org:
                LinkAnchor anchor = new LinkAnchor(
                    id: directive.value,
                    directive: directive,
                    sourceDocId: processor.currentDocId)

                processor.linkAnchors[anchor.id] = anchor
                break;
            default:
                break // do nothing
            } }

    protected void parse(CodeBlock codeBlock) {} // nothing to do
    protected void parse(DocText docText) {} // nothing to do

    protected String emit(SourceFile sourceFile) {
        StringBuilder sb = new StringBuilder()


        // Create the HTML file head
        sb.append(
"""<!DOCTYPE html>
<html>
    <head>
        <title>${processor.currentDocId}</title>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        <link rel="stylesheet" media="all" href="jlp.css"/>
    </head>
    <body>
        <div id="container">
            <table cellpadding="0" cellspacing="0">
                <thead><tr>
                    <th class="docs"><h1>${processor.currentDocId}</h1></th>
                    <th class="code"/>
                </tr></thead>
                <tbody>""")

        // Emit the document to Markdown
        sourceFile.blocks.each { block -> sb.append(emit(block)) }

        // Create the HTML file foot
        sb.append(
"""                </tbody>
            </table>
        </div>
    </body>
</html>""")

        return sb.toString() }

    protected String emit(Block block) {
        StringBuilder sb = new StringBuilder()

        // Look for an `@org` directive in the `Block`
        Directive orgDir = block.docBlock.directives.find {
            it.type == DirectiveType.Org }

        // Create the `tr` that will hold the `Block`
        if (orgDir) { sb.append("\n<tr id='${orgDir.value}'>") }
        else        { sb.append("<tr>") }

        // Create the `td` for the documentation.
        sb.append('\n<td class="docs">')
        sb.append(emit(block.docBlock))
        sb.append('</td>')

        // Create the `td` for the `CodeBlock`
        sb.append('\n<td class="code">')
        sb.append(emit(block.codeBlock))
        sb.append('</td>')

        // Close the table row.
        sb.append('</tr>') }

    protected String emit(DocBlock docBlock) {
        // Create a queue for the doc block elements, we are going to 
        // sort them by type and line number
        List emitQueue 

        // Later we will need a string builder to hold our result.
        StringBuilder sb

        // Add all the directives
        emitQueue = docBlock.directives.collect { directive ->
            def queueItem = [lineNumber: directive.lineNumber, value: directive]
            switch(directive.type) {
                case DirectiveType.Api:         queueItem.priority = 50; break
                case DirectiveType.Author:      queueItem.priority = 10; break
                case DirectiveType.Copyright:   queueItem.priority = 11; break
                case DirectiveType.Example:     queueItem.priority = 50; break
                case DirectiveType.Org:         queueItem.priority =  0; break }
            
            return queueItem }

        // Add all the doc text blocks
        emitQueue.addAll(docBlock.docTexts.collect { docText ->
            [lineNumber: docText.lineNumber, priority: 50, value: docText] })
                        

        // Sort the emit queue by priority, then line number.
        emitQueue.sort(
            {i1, i2 -> i1.priority != i2.priority ?
                        i1.priority - i2.priority :
                        i1.lineNumber - i2.lineNumber} as Comparator)
    
        // Finally, we want to treat the whole block as one markdown chunk, so
        // we will concatenate the values in the emit queue and then process
        // the whole block once
        sb = new StringBuilder()
        emitQueue.each { queueItem -> sb.append(emit(queueItem.value)) } 

        return processMarkdown(sb.toString())
    }

    protected String emit(CodeBlock codeBlock) {
        def codeLines

        // Collect the lines into an array.
        codeLines = codeBlock.lines.collect { lineNumber, line ->
            [lineNumber, line] }

        // Sort by line number.
        codeLines.sort({ i1, i2 -> i1[0] - i2[0] } as Comparator)

        codeLines = codeLines.collect { arr -> arr[1] }

        // write out the lines in a <pre><code> block
        return "<pre><code>${codeLines.join('')}</code></pre>" }

    protected String emit(DocText docText) { return docText.value }

    protected String emit(Directive directive) {
        switch(directive.type) {

            // An `@api` directive is immediately processed and wrapped in a
            // div (we need to process this now because Markdown does not
            // process input inside HTML elements).
            case DirectiveType.Api:
                return "<div class='api'>" +
                    processMarkdown(directive.value) + "</div>\n"

            // An `@author` directive is turned into a definition list.
            case DirectiveType.Author:
                return "Author\n:   ${directive.value}\n"

            case DirectiveType.Copyright:
                return "\n&copy; ${directive.value}\n"

            // An `@example` directive is returned as is
            case DirectiveType.Example:
                return directive.value

            // An `@org` directive is ignored.
            case DirectiveType.Org: return "" }
    }

    protected String processMarkdown(String markdown) {
        // convert to HTML from Markdown
        String html = pegdown.markdownToHtml(markdown)

        // replace internal `jlp://` links with actual links based on`@org`
        // references
        html = html.replaceAll(/jlp:\/\/([^\s"]+)/) { wholeMatch, linkId ->

            // Get the org data stored for this org id.
            def link = processor.linkAnchors[linkId]
            String newLink

            if (!link) {
                // We do not have any reference to this id.
                /* TODO: log error */
                newLink = "broken_link(${linkId})" }

            // This link points to a location in this document.
            else if (processor.currentDocId == link.sourceDocId) {
                newLink = "#$linkId" }

            // The link should point to a different document.
            else {
                TargetDoc thisDoc = processor.currentDoc
                TargetDoc linkDoc = processor.docs[link.sourceDocId]

                String pathToLinkedDoc = processor.getRelativeFilepath(
                    thisDoc.sourceFile.parentFile,
                    linkDoc.sourceFile)

                // The target document may not be in the same directory
                // as us, backtrack to the (relative) top of our directory
                // structure.
                newLink = pathToLinkedDoc + ".html#${linkId}" }
                
            return newLink }

        return html;
    }
}
